import redis
from os import getenv


class RedisDB:
    __slots__ = ['redis']

    __redis_host = getenv('REDIS_HOST') or '0.0.0.0'
    __redis_password = getenv('REDIS_PASSWORD') or 'redispass'
    __redis_port = getenv('REDIS_PORT') or 6379

    def __init__(self):
        self.__connect()

    def __connect(self):
        try:
            self.redis = redis.Redis(host=self.__redis_host, password=self.__redis_password, port=self.__redis_port)

        except Exception as ex:
            print(ex)



from json import dumps
from pwx_db import PostgresDB
from datetime import datetime

db = PostgresDB()


class EntitySync:

    __table_name__ = ''

    def __init__(self, name):
        self.__table_name__ = name

    def __sync_attrs__(self):
        try:
            for attr in db.select_columns(self.__table_name__):
                setattr(self, attr[0], None)

        except NameError as ne:
            raise NameError(ne.args[0])

    def __get_attrs__(self):
        return [attr for attr in self.__dict__.keys() if not self.__is_table_name(attr)]

    def __set_values__(self, values):
        attrs = self.__get_attrs__()
        for i in range(len(attrs)):
            self.__setattr__(attrs[i], values[i])

    def __to_json__(self, values=None):
        self.__set_values__(values)

        json = {}

        for attr, value in self.__dict__.items():
            if not self.__is_table_name(attr):

                if isinstance(value, datetime):
                    value = str(value)

                json.update({attr: value})

        return json['id'], dumps(json)

    @staticmethod
    def __is_table_name(attr):
        if not attr.startswith('__t'):
            return False

        return True



from re import sub
from pwx_db import PostgresDB, Native


class PersistSync:

    __slots__ = ['__table_name', '__columns', '__values', '__sql_delete', '__sql_insert', '__ids']

    __db = PostgresDB()

    def __init__(self, table_name):
        self.__table_name = table_name
        self.__ids = []

    def __init(self, columns, values, only_insert):
        self.__columns = columns
        self.__values = values

        if not only_insert:
            self.__get_ids()

    def persist(self, columns, values, many=False, only_insert=False):
        self.__init(columns, values, only_insert)
        self.__prepare_sql(many, only_insert)
        self.__execute(only_insert)
        self.__ids = []

    def __get_ids(self):
        if isinstance(self.__values[0], list):
            for value in self.__values:
                self.__ids.append(int(value[0]))  # first position always be id
        else:
            self.__ids.append(self.__values[0])

    def __prepare_sql(self, many, only_insert):
        if not only_insert:
            self.__sql_delete = Native.delete_by_id(table_name=self.__table_name, ident=self.__get_string(self.__ids))

        self.__sql_insert = Native.insert(table_name=self.__table_name,
                                          colunms=self.__get_string(self.__columns, is_columns=True),
                                          values=self.__get_string(self.__values, many))

    def __get_string(self, values, many=False, is_columns=False):
        if many:
            return self.__many_values(values, is_columns)

        return self.__to_string(values, is_columns)

    def __to_string(self, values, is_columns):
        return f"({', '.join([self.__convert(value, is_columns) for value in values])})"

    @staticmethod
    def __convert(value, is_columns):
        if is_columns:
            return str(value)

        elif isinstance(value, bool):
            return "'" + f'{str(value).lower()}' + "'"

        elif isinstance(value, int) or isinstance(value, float):
            return str(value)

        elif isinstance(value, str):
            return "'" + f'{value.lower()}' + "'"

        elif not value:
            return 'null'

    def __many_values(self, values, is_columns):
        val = ''
        for value in values:
            val += f'{self.__to_string(value, is_columns)},'

        return sub('[,]$', ';', val)

    def __execute(self, only_insert):
        if not only_insert:
            self.__execute_delete()

        self.__execute_insert()

    def __execute_delete(self):
        self.__db.execute_query(sql=self.__sql_delete)

    def __execute_insert(self):
        self.__db.execute_query(sql=self.__sql_insert)



from json import loads, dumps
from .pwx_redis import RedisDB


class RedisPWX:

    __slots__ = []

    redis = RedisDB().redis

    def __init__(self):
        pass

    def add(self, collection, key, value):
        self.redis.hset(name=collection, key=key, value=value)

    def add_many(self, collection, values):
        self.redis.hmset(name=collection, mapping=values)

    def get_by_key(self, collection, key):
        result = self.redis.hget(name=collection, key=key)

        if result:
            return self.__loads_result(result)

        return None

    def get_many_by_keys(self, collection, values=None):
        if not isinstance(values, list):
            raise TypeError

        objects = []

        for value in values:
            obj = self.get_by_key(collection, key=value)

            if not obj:
                return None

            objects.append(obj)

        return objects

    def get_by_value(self, collection, attr=None, value=None):
        all_result = self.get_all(collection)

        if all_result:
            for k in all_result:
                values = self.__loads_result(all_result[k])
                if values[attr].__eq__(value):
                    return int(k)

            return None

        else:
            return None

    def get_many_id_by_value(self, collection, attr=None, value=None):
        all_result = self.get_all(collection)

        if all_result:
            result = []

            for k in all_result:
                values = self.__loads_result(all_result[k])
                if values[attr].__eq__(value):
                    result.append(int(k))

            if result:
                return result

            return None

        else:
            return None

    def get_by_values(self, collection, attrs, values):
        if not isinstance(attrs, list) or not isinstance(values, list):
            raise TypeError

        if not len(attrs).__eq__(len(values)):
            raise AssertionError

        all_result = self.get_all(collection=collection)

        if all_result:
            for k in all_result:
                obj_values = self.__loads_result(all_result[k])
                is_return = True

                for i in range(len(attrs)):
                    if not obj_values[attrs[i]].__eq__(values[i]):
                        is_return = False
                        break

                if is_return:
                    return int(k)

            return None

        else:
            return None

    def get_all(self, collection):
        return self.redis.hgetall(name=collection)

    def get_entity_by_value(self, collection, attr=None, value=None):
        all_result = self.get_all(collection)

        if all_result:
            for k in all_result:
                values = self.__loads_result(all_result[k])
                if values[attr] == value:
                    return self.get_by_key(collection=collection, key=int(k))

            return None

        else:
            return None

    def get_entity_by_values(self, collection, attrs, values):
        if not isinstance(attrs, list) or not isinstance(values, list):
            raise TypeError

        if not len(attrs).__eq__(len(values)):
            raise AssertionError

        all_result = self.get_all(collection=collection)

        if all_result:
            for k in all_result:
                obj_values = self.__loads_result(all_result[k])
                is_return = True

                for i in range(len(attrs)):
                    if not obj_values[attrs[i]] == values[i]:
                        is_return = False
                        break

                if is_return:
                    return self.get_by_key(collection=collection, key=int(k))

            return None

        else:
            return None

    def get_many_entities_by_value(self, collection, attr=None, value=None):
        all_result = self.get_all(collection)

        if all_result:
            result = []

            for k in all_result:
                values = self.__loads_result(all_result[k])
                if values[attr].__eq__(value):
                    result.append(self.get_by_key(collection=collection, key=int(k)))

            if result:
                return result

            return None

        else:
            return None

    def get_many_entities_by_values(self, collection, attrs, values):
        if not isinstance(attrs, list) or not isinstance(values, list):
            raise TypeError

        if not len(attrs).__eq__(len(values)):
            raise AssertionError

        all_result = self.get_all(collection=collection)

        if all_result:
            result = []

            for k in all_result:
                is_return = True
                obj_values = self.__loads_result(all_result[k])

                for i in range(len(attrs)):
                    if not isinstance(values[i], list):
                        if not obj_values[attrs[i]] == values[i]:
                            is_return = False
                            break

                    else:
                        for j in values[i]:
                            if obj_values[attrs[i]] == j:
                                is_return = True
                                break
                            else:
                                is_return = False

                if is_return:
                    result.append(obj_values)

            if result:
                return result

            return None

        return None

    def update(self, collection, key, value):
        self.redis.hset(name=collection, key=key, value=dumps(value))

    def delete(self, collection, key):
        self.redis.hdel(collection, key)

    def delete_many(self, collection, *keys):
        self.redis.hdel(collection, *keys)

    @staticmethod
    def __loads_result(result):
        return loads(result)



from .enums import *
from .db import *
from .services import *
from .sync import *

__all__ = ['InstanceStateIntEnum', 'StepTypeEnum', 'TimeoutMessagesEnum', 'PostgresDB', 'Native', 'RedisPWX',
           'BaseService', 'EntitySync', 'PersistSync', 'InstanceStatusIntEnum', 'ORMSync']



import psycopg2
from os import getenv


class PostgresDB:
    __slots__ = ['__connection', '__cursor']

    __db_name = getenv('DB_NAME')
    __db_user = getenv('DB_USER')
    __db_password = getenv('DB_PASS')
    __db_host = getenv('DB_HOST')

    def __init__(self):
        self.__connection = None
        self.__cursor = None

    def __connect(self):
        try:
            self.__connection = psycopg2.connect(dbname=self.__db_name,
                                                 user=self.__db_user,
                                                 password=self.__db_password,
                                                 host=self.__db_host)

            self.__cursor = self.__connection.cursor()

        except Exception as ex:
            print(ex)

    def __set_cursor(self):
        self.__cursor = self.__connection.cursor()

    def close_cursor(self):
        if self.__cursor and not self.__cursor.closed:
            self.__cursor.close()

    @property
    def cursor(self):
        self.__check_connection()
        if self.__cursor.closed:
            self.__set_cursor()

        return self.__cursor

    def select_columns(self, table_name):
        try:
            self.__check_connection()

            self.cursor.execute(self.__get_native_sql(table_name))
            result = self.cursor.fetchall()

            if result:
                return result

            raise NameError(f'Table with name [{table_name if table_name else "empty name"}] not exists!')

        except Exception as ex:
            raise ex

        finally:
            self.close_cursor()
            self.__close_connection()

    def execute_query_all(self, sql):
        try:
            self.__check_connection()

            self.cursor.execute(sql)
            result = self.cursor.fetchall()

            if result:
                return result

            return None

        except Exception as ex:
            print(ex, flush=True)
            return None

        finally:
            self.__close_connection()

    def execute_query_one(self, sql):
        try:
            self.__check_connection()

            self.cursor.execute(sql)
            result = self.cursor.fetchone()

            if result:
                return result

            return None

        except Exception as ex:
            print(ex, flush=True)
            return None

        finally:
            self.__close_connection()

    def execute_query(self, sql):
        self.__check_connection()

        try:
            self.cursor.execute(sql)
            self.__connection.commit()

        except Exception as ex:
            print(f'[ERRO SQL] ::: {ex}')
            self.__connection.rollback()

        self.__close_connection()

    def __check_connection(self):
        if not self.__connection or self.__connection.closed:
            self.__connect()

    def __close_connection(self):
        if not self.__connection.closed:
            self.__connection.close()
            self.close_cursor()

    @staticmethod
    def __get_native_sql(table_name):
        return f"SELECT column_name FROM information_schema.columns where table_name = '{table_name}'"



from setuptools import setup, find_packages

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pwx-db-connection',
    version='1.3.0',
    url='',
    author='Guilherme Rosa Koerich',
    author_email='guilherme@pwx.cloud',
    description='Connection to PostgreSQL and Redis - PWX',
    keywords='connection redis postgres pwx',
    packages=find_packages(exclude=['tests']),
    long_description=long_description,
    long_description_content_type='text/markdown',
    zip_safe=False,
    install_requires=['redis==3.3.11', 'psycopg2-binary==2.8.4'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>3.4',
)



from json import dumps
from datetime import datetime


class ORMSync:
    __slots__ = ['__keys', '__json', '__id', '__attrs_to_ignore', '__attrs_to_id']

    __default_ignore = ['__', '_', 'query', 'metadata']

    def __init__(self):
        self.__id = ''
        self.__attrs_to_id = None
        self.__json = {}
        self.__attrs_to_ignore = []

    def __prepare(self, attrs_to_ignore=None, attrs_to_id=None):
        self.__attrs_to_ignore = attrs_to_ignore or []
        self.__attrs_to_id = attrs_to_id or None
        self.__id = ''

    def to_json(self, obj, attrs_to_ignore=None, attrs_to_id=None):
        self.__prepare(attrs_to_ignore, attrs_to_id)
        self.__set_keys(obj)
        self.__generate_json(obj)

        return self.__id, self.__json

    def __set_keys(self, obj):
        self.__keys = [key for key in dir(obj) if not self.__ignore_keys(key, obj)]

    def __ignore_keys(self, key, obj):
        for ignore in self.__default_ignore:
            if key.startswith(ignore):
                return True

        return True if callable(getattr(obj, key)) or self.__attrs_to_ignore.__contains__(key) else False

    def __generate_json(self, obj):
        self.__clear_json()

        for key in self.__keys:
            if isinstance(obj.__getattribute__(key), list):
                continue

            elif isinstance(obj.__getattribute__(key), datetime):
                self.__json.update({key: str(obj.__getattribute__(key))})

            else:
                self.__json.update({key: obj.__getattribute__(key)})

        self.__set_id()
        self.__dumps_json()

    def __dumps_json(self):
        self.__json = dumps(self.__json)

    def __set_id(self):
        if self.__attrs_to_id:
            for att in self.__attrs_to_id:
                if not self.__id:
                    self.__id = self.__json.get(att)

                else:
                    self.__id = f'{self.__id}_{self.__json.get(att)}'

        else:
            self.__id = self.__json.get('id')

    def __clear_json(self):
        if self.__json:
            self.__json = {}
